@@ -1,6 +1,6 @@
 #!/usr/bin/env python
 #
-# Copyright 2010 Facebook
+# Copyright 2010-present Facebook
 #
 # Licensed under the Apache License, Version 2.0 (the "License"); you may
 # not use this file except in compliance with the License. You may obtain
@@ -26,7 +26,19 @@ import stat
 import time
 import types
 import urllib2
+import urllib
 import webbrowser
+import StringIO
+
+poster_is_available = False
+try:
+    # try to use poster if it is available
+    import poster.streaminghttp
+    import poster.encode
+    poster.streaminghttp.register_openers()
+    poster_is_available = True
+except ImportError:
+    pass # we can live without this.
 
 from urlparse import urlparse, parse_qs
 from urllib import urlencode
@@ -38,31 +50,35 @@ ACCESS_TOKEN = None
 ACCESS_TOKEN_FILE = '.fb_access_token'
 AUTH_SCOPE = []
 
+AUTH_SUCCESS_HTML = """
+You have successfully logged in to facebook with fbconsole.
+You can close this window now.
+"""
+
 __all__ = [
     'help',
     'authenticate',
     'logout',
+    'graph_url',
     'get',
     'post',
     'delete',
     'shell',
     'fql',
+    'iter_pages',
     'APP_ID',
     'SERVER_PORT',
     'ACCESS_TOKEN',
     'AUTH_SCOPE',
     'ACCESS_TOKEN_FILE']
 
-def _get_url(path, args=None, graph=True):
+def _get_url(path, args=None):
     args = args or {}
     if ACCESS_TOKEN:
         args['access_token'] = ACCESS_TOKEN
-    if graph:
-        subdomain = 'graph'
-        if '/videos' in path:
-            subdomain = 'graph-video'
-    else:
-        subdomain = 'graph' if graph else 'api'
+    subdomain = 'graph'
+    if '/videos' in path:
+        subdomain = 'graph-video'
     if 'access_token' in args or 'client_secret' in args:
         endpoint = "https://%s.facebook.com" % subdomain
     else:
@@ -140,8 +156,7 @@ class _RequestHandler(BaseHTTPServer.BaseHTTPRequestHandler):
                 else:
                     data['expires_at'] = int(time.time()+int(expiration))
             open(ACCESS_TOKEN_FILE,'w').write(json.dumps(data))
-            self.wfile.write("You have successfully logged in to facebook with fbconsole. "
-                             "You can close this window now.")
+            self.wfile.write(AUTH_SUCCESS_HTML)
         else:
             self.wfile.write('<html><head>'
                              '<script>location = "?"+location.hash.slice(1);</script>'
@@ -156,6 +171,7 @@ help() - display this help message
 authenticate() - authenticate with facebook.
 logout() - Remove the cached access token, forcing authenticate() to get a new
            access token
+graph_url(path, params) - get the full url to a graph api path
 get(path, params) - call the graph api with the given path and query parameters
 post(path, data) - post data to the graph api with the given path
 delete(path, params) - send a delete request
@@ -166,7 +182,7 @@ def authenticate():
     """Authenticate with facebook so you can make api calls that require auth.
 
     Alternatively you can just set the ACCESS_TOKEN global variable in this
-    module to an access token you get from facebook.
+    module to set an access token you get from facebook.
 
     If you want to request certain permissions, set the AUTH_SCOPE global
     variable to the list of permissions you want.
@@ -182,7 +198,6 @@ def authenticate():
             needs_auth = False
 
     if needs_auth:
-        print "Logging you in to facebook..."
         webbrowser.open('https://www.facebook.com/dialog/oauth?' +
                         urlencode({'client_id':APP_ID,
                                    'redirect_uri':'http://127.0.0.1:%s/' % SERVER_PORT,
@@ -198,31 +213,82 @@ def logout():
     if os.path.exists(ACCESS_TOKEN_FILE):
         os.remove(ACCESS_TOKEN_FILE)
 
+def graph_url(path, params=None):
+    """Get the full url to the graph api for the given path and query args.
+
+    This is useful if you want to use your own method of making http requests or
+    are not interested in the json parsing that occurs by default. For example,
+    download a large profile picture of Mark Zuckerberg:
+
+      >>> url = graph_url('/zuck/picture', {"type":"large"})
+      >>> filename, response = urllib.urlretrieve(url, 'mark.jpg')
+
+    """
+    return _get_url(path, args=params)
+
 def get(path, params=None):
     """Send a GET request to the graph api.
 
     For example:
 
-      >>> get('/me')
-      >>> get('/me', {'fields':'id,name'})
+      >>> user = get('/me')
+      >>> print user['first_name']
+      David
+      >>> short_user = get('/me', {'fields':'id,first_name'})
+      >>> print short_user['id'], short_user['first_name']
+      100003169144448 David
 
     """
     return json.load(urllib2.urlopen(_get_url(path, args=params)))
 
+def iter_pages(json_response):
+    """Iterate over multiple pages of data.
+
+    The graph api can return a lot of data, but will only return a limited
+    amount of data in a single request.  To get more data, you must query for it
+    explicitly using paging.  This function will do automatic paging in the form
+    of an iterator.  For example to print the id of every photo tagged with the
+    logged in user:
+
+      >>> total = 0
+      >>> for photo in iter_pages(get('/19292868552/feed', {'limit':2})):
+      ...     total += 1
+      ...     print "There are at least", total, "feed stories"
+      ...     if total > 4: break
+      There are at least 1 feed stories
+      There are at least 2 feed stories
+      There are at least 3 feed stories
+      There are at least 4 feed stories
+      There are at least 5 feed stories
+
+    """
+    while len(json_response.get('data','')):
+        for item in json_response['data']:
+            yield item
+        next_url = json_response['paging']['next']
+        json_response = json.load(urllib2.urlopen(next_url))
+
 def post(path, params=None):
     """Send a POST request to the graph api.
 
     You can also upload files using this function.  For example:
 
-      >>> post('/me/photos',
+      >>> photo_id = post('/me/photos',
       ...            {'name': 'My Photo',
-      ...             'source': open("myphoto.jpg")})
+      ...             'source': open("icon.gif")})['id']
+      >>> print get('/'+photo_id)['name']
+      My Photo
 
     """
-    opener = urllib2.build_opener(
-        urllib2.HTTPCookieProcessor(cookielib.CookieJar()),
-        _MultipartPostHandler)
-    return json.load(opener.open(_get_url(path), params))
+    if poster_is_available:
+        data, headers = poster.encode.multipart_encode(params)
+        request = urllib2.Request(_get_url(path), data, headers)
+        return json.load(urllib2.urlopen(request))
+    else:
+        opener = urllib2.build_opener(
+            urllib2.HTTPCookieProcessor(cookielib.CookieJar()),
+            _MultipartPostHandler)
+        return json.load(opener.open(_get_url(path), params))
 
 def delete(path, params=None):
     """Send a DELETE request to the graph api.
@@ -231,6 +297,7 @@ def delete(path, params=None):
 
       >>> msg_id = post('/me/feed', {'message':'hello world'})['id']
       >>> delete('/'+msg_id)
+      True
 
     """
     if not params:
@@ -244,12 +311,11 @@ def fql(query):
     For example:
 
       >>> fql('SELECT name FROM user WHERE uid = me()')
+      [{u'name': u'David Amcafiaddddh Yangstein'}]
 
     """
-    url = _get_url('/method/fql.query',
-                   args={'query': query, 'format': 'json'},
-                   graph=False)
-    return json.load(urllib2.urlopen(url))
+    url = _get_url('/fql', args={'q': query})
+    return json.load(urllib2.urlopen(url))['data']
 
 INTRO_MESSAGE = '''\
   __ _                                _
@@ -275,5 +341,12 @@ def shell():
         import code
         code.InteractiveConsole(globals()).interact(INTRO_MESSAGE)
 
+
+def test_suite():
+    import doctest
+    global ACCESS_TOKEN
+    ACCESS_TOKEN = 'AAACjeiZB6FgIBAB8eZABg7So8ALDisFLugfIJSZCg3FEDRy82yEmdXYYfNvdv2kWVMWxaJgWqqVMPtG5v5n4lMG5VXmZBZBykQkeluhpFPQZDZD'
+    return doctest.DocTestSuite()
+
 if __name__ == '__main__':
     shell()
